# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from ...config import config
from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Circle"

def parse_circle(in_row):
    if in_row[2] in ("deposit", "internal_switch_currency", "switch_currency"):
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[7].strip('£').split(' ')[0],
                                    buy_asset=in_row[8],
                                    sell_quantity=in_row[5].strip('£').split(' ')[0],
                                    sell_asset=in_row[6],
                                    wallet=WALLET)
    elif in_row[2] == "spend":
        return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                    DataParser.parse_timestamp(in_row[0]),
                                    sell_quantity=in_row[5].strip('£').split(' ')[0],
                                    sell_asset=in_row[6],
                                    sell_value=in_row[7].strip('£') if in_row[8] == config.CCY \
                                                                    else None,
                                    wallet=WALLET)
    elif in_row[2] == "receive":
        return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[7].strip('£').split(' ')[0],
                                    buy_asset=in_row[8],
                                    buy_value=in_row[5].strip('£') if in_row[6] == config.CCY \
                                                                   else None,
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised Transaction Type: " + in_row[2])

DataParser(DataParser.TYPE_EXCHANGE,
           "Circle",
           ['Date', ' Reference ID', ' Transaction Type', ' From Account', ' To Account',
            ' From Amount', ' From Currency', ' To Amount', ' To Currency', ' Status'],
           row_handler=parse_circle)



# -*- coding: utf-8 -*-
# Cryptoasset accounting, auditing and UK tax calculations (Capital Gains/Income Tax)
# (c) Nano Nano Ltd 2019

import logging
import argparse
import io
import sys

from .version import __version__
from .config import config
from .transactions import load_transaction_records, TransactionHistory
from .audit import audit_transactions
from .price.valueasset import ValueAsset
from .tax import TaxCalculator

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='[%(asctime)s.%(msecs)03d] %(levelname)s -- : %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')
log = logging.getLogger()

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("filename",
                        type=str,
                        nargs='?',
                        help="filename of transaction records, "
                             "or can read from standard input")
    parser.add_argument("-ty",
                        "--taxyear",
                        type=int,
                        help="tax year")
    parser.add_argument("-s",
                        "--skipaudit",
                        action='store_true',
                        help="skip auditing of transactions")
    parser.add_argument("-d",
                        "--debug",
                        action='store_true',
                        help="enabled debug logging")
    parser.add_argument("-v",
                        "--version",
                        action='version',
                        version='%(prog)s {version}'.format(version=__version__))

    config.args = parser.parse_args()
    config.args.nocache = False

    if config.args.debug:
        log.setLevel(logging.DEBUG)
        config.output_config(parser.prog)

    if config.args.filename:
        with io.open(config.args.filename, newline='', encoding='utf-8') as import_file:
            transaction_records = load_transaction_records(import_file)
            import_file.close()
    else:
        transaction_records = load_transaction_records(sys.stdin)

    if not config.args.skipaudit:
        audit_transactions(sorted(transaction_records))

    transaction_history = TransactionHistory(transaction_records)
    value_asset = ValueAsset()
    transactions = transaction_history.split_transaction_records(value_asset)

    tax = TaxCalculator(transactions)
    tax.pool_same_day()
    tax.match(tax.DISPOSAL_SAME_DAY)
    tax.match(tax.DISPOSAL_BED_AND_BREAKFAST)

    if config.args.debug:
        tax.output_transactions()

    tax.process_unmatched()
    tax.process_income()

    if config.args.taxyear:
        tax.report_capital_gains(config.args.taxyear)
        tax.report_income(config.args.taxyear)
    else:
        # Output for all years
        for year in sorted(tax.tax_events):
            tax.report_capital_gains(year)
            tax.report_income(year)

        tax.report_holdings(value_asset)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal, ROUND_UP

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Poloniex"

PRECISION = Decimal('0.00000000')

def parse_poloniex_trades(in_row):
    if in_row[3] == "Buy":
        fee_quantity = Decimal(in_row[5]) - Decimal(in_row[5]) * \
                (Decimal(in_row[7].replace('%', '')) / Decimal(100))
        fee_quantity = fee_quantity.quantize(PRECISION, rounding=ROUND_UP)
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[5],
                                    buy_asset=in_row[1].split('/')[0],
                                    sell_quantity=in_row[6],
                                    sell_asset=in_row[1].split('/')[1],
                                    fee_quantity=Decimal(in_row[5]) - fee_quantity,
                                    fee_asset=in_row[1].split('/')[0],
                                    wallet=WALLET)
    elif in_row[3] == "Sell":
        fee_quantity = Decimal(in_row[6]) - Decimal(in_row[6]) * \
                (Decimal(in_row[7].replace('%', '')) / Decimal(100))
        fee_quantity = fee_quantity.quantize(PRECISION, rounding=ROUND_UP)
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[6],
                                    buy_asset=in_row[1].split('/')[1],
                                    sell_quantity=in_row[5],
                                    sell_asset=in_row[1].split('/')[0],
                                    fee_quantity=Decimal(in_row[6]) - fee_quantity,
                                    fee_asset=in_row[1].split('/')[1],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised Type: " + in_row[3])

def parse_poloniex_deposits_withdrawals(in_row):
    if "COMPLETE:" in in_row[4]:
        return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                    DataParser.parse_timestamp(in_row[0]),
                                    sell_quantity=in_row[2],
                                    sell_asset=in_row[1],
                                    wallet=WALLET)
    else:
        return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[2],
                                    buy_asset=in_row[1],
                                    wallet=WALLET)

def parse_poloniex_withdrawals(in_row, _):
    return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                DataParser.parse_timestamp(in_row[0]),
                                sell_quantity=Decimal(in_row[2]) - Decimal(in_row[3]),
                                sell_asset=in_row[1],
                                fee_quantity=in_row[3],
                                fee_asset=in_row[1],
                                wallet=WALLET)

DataParser(DataParser.TYPE_EXCHANGE,
           "Poloniex Trades",
           ['Date', 'Market', 'Category', 'Type', 'Price', 'Amount', 'Total', 'Fee', 'Order Number',
            'Base Total Less Fee', 'Quote Total Less Fee'],
           row_handler=parse_poloniex_trades)

DataParser(DataParser.TYPE_EXCHANGE,
           "Poloniex Deposits/Withdrawals",
           ['Date', 'Currency', 'Amount', 'Address', 'Status'],
           row_handler=parse_poloniex_deposits_withdrawals)

DataParser(DataParser.TYPE_EXCHANGE,
           "Poloniex Withdrawals",
           ['Date', 'Currency', 'Amount', 'Fee Deducted', 'Amount - Fee', 'Address', 'Status'],
           row_handler=parse_poloniex_withdrawals)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import copy
from decimal import Decimal
from datetime import datetime, timedelta

from .config import config
from .transactions import Buy, Sell
from .holdings import Holdings

log = logging.getLogger()

class TaxCalculator(object):
    DISPOSAL_SAME_DAY = 'Same Day'
    DISPOSAL_BED_AND_BREAKFAST = 'Bed & Breakfast'
    DISPOSAL_SECTION_104 = 'Section 104'

    INCOME_TYPES = (Buy.TYPE_MINING, Buy.TYPE_INCOME)
    TRANSFER_TYPES = (Buy.TYPE_DEPOSIT, Sell.TYPE_WITHDRAWAL)

    CGT_RATE = 20
    ANNUAL_ALLOWANCE = {2010: 10100, 2011: 10100, 2012: 10600, 2013: 10600, 2014: 10900,
                        2015: 11000, 2016: 11100, 2017: 11100, 2018: 11300, 2019: 11700,
                        2020: 12000}

    PRECISION = Decimal('0.00')

    def __init__(self, transactions):
        self.transactions = transactions
        self.buys_ordered = []
        self.sells_ordered = []
        self.other_transactions = []

        self.tax_events = {}
        self.holdings = {}

    def pool_same_day(self):
        transactions = copy.deepcopy(self.transactions)
        buy_transactions = {}
        sell_transactions = {}

        log.debug("==POOL SAME DAY TRANSACTIONS==")
        for t in transactions:
            if isinstance(t, Buy) and t.is_acquisition():
                if (t.asset, t.timestamp.date()) not in buy_transactions:
                    buy_transactions[(t.asset, t.timestamp.date())] = t
                else:
                    buy_transactions[(t.asset, t.timestamp.date())] += t
            elif isinstance(t, Sell) and t.is_disposal():
                if (t.asset, t.timestamp.date()) not in sell_transactions:
                    sell_transactions[(t.asset, t.timestamp.date())] = t
                else:
                    sell_transactions[(t.asset, t.timestamp.date())] += t
            else:
                self.other_transactions.append(t)

        self.sells_ordered = sorted(sell_transactions.values())
        self.buys_ordered = sorted(buy_transactions.values())

        if config.args.debug:
            for t in sorted(self.sells_ordered + self.buys_ordered + self.other_transactions):
                log.debug(t)
                if len(t.pooled) > 1:
                    for tp in t.pooled:
                        log.debug("  %s", tp)

            log.debug("Total Transactions(Pooled)=%s",
                      len(self.sells_ordered + self.buys_ordered + self.other_transactions))

    def match(self, rule):
        log.debug("==MATCH %s TRANSACTIONS==", rule.upper())
        sell_index = buy_index = 0

        if not self.buys_ordered:
            return

        while sell_index < len(self.sells_ordered):
            s = self.sells_ordered[sell_index]
            b = self.buys_ordered[buy_index]

            if (not s.matched and not b.matched and s.asset == b.asset and
                    self._rule_match(s.timestamp, b.timestamp, rule)):
                if config.args.debug:
                    log.debug(s)
                    if rule == self.DISPOSAL_BED_AND_BREAKFAST:
                        log.debug("%s (%s, %s days)",
                                  b, rule, (b.timestamp.date() - s.timestamp.date()).days)
                    else:
                        log.debug("%s (%s)", b, rule)

                if b.buy_quantity > s.sell_quantity:
                    b_remainder = b.split_buy(s.sell_quantity)
                    self.buys_ordered.insert(buy_index + 1, b_remainder)
                elif s.sell_quantity > b.buy_quantity:
                    s_remainder = s.split_sell(b.buy_quantity)
                    self.sells_ordered.insert(sell_index + 1, s_remainder)

                s.matched = b.matched = True
                tax_event = TaxCapitalGains(rule, b, s, b.cost)
                self.tax_events[self._which_tax_year(tax_event.date)].append(tax_event)

                # Find next sell
                sell_index += 1
                buy_index = 0
            else:
                buy_index += 1
                if buy_index >= len(self.buys_ordered):
                    sell_index += 1
                    buy_index = 0

    def output_transactions(self):
        log.debug("==UPDATED TRANSACTIONS==")
        for t in sorted(self.buys_ordered +
                        self.sells_ordered +
                        self.other_transactions):
            log.debug(t)

        log.debug("Total Transactions=%s",
                  len(self.sells_ordered + self.buys_ordered + self.other_transactions))

    def _rule_match(self, s_timestamp, b_timestamp, rule):
        if rule == self.DISPOSAL_SAME_DAY:
            return b_timestamp.date() == s_timestamp.date()
        elif rule == self.DISPOSAL_BED_AND_BREAKFAST:
            return (s_timestamp.date() < b_timestamp.date() and
                    b_timestamp.date() <= s_timestamp.date() + timedelta(days=30))
        else:
            raise Exception

    def process_unmatched(self):
        log.debug("==PROCESS UNMATCHED TRANSACTIONS==")
        unmatched_transactions = sorted([t for t in self.buys_ordered +
                                         self.sells_ordered +
                                         self.other_transactions if t.matched is False])

        for t in unmatched_transactions:
            if config.args.debug:
                if isinstance(t, Sell) and t.is_disposal():
                    log.debug("%s (Disposal)", t)
                else:
                    log.debug(t)

            if isinstance(t, Buy):
                self._add_tokens(t)
            elif isinstance(t, Sell):
                self._subtract_tokens(t)

    def _add_tokens(self, t):
        if t.asset not in self.holdings:
            self.holdings[t.asset] = Holdings(t.asset)

        if t.t_type in self.TRANSFER_TYPES:
            if config.transfers_include:
                # !IMPORTANT! - Make sure no disposal event occurs between a Withdrawal and a
                #   Deposit (of the same asset) otherwise the average cost basis would be incorrect
                cost = Decimal(0)
            else:
                return
        else:
            cost = t.cost

        self.holdings[t.asset].add_tokens(t.buy_quantity, cost)

    def _subtract_tokens(self, t):
        if t.asset not in self.holdings:
            self.holdings[t.asset] = Holdings(t.asset)

        if t.t_type in self.TRANSFER_TYPES:
            if config.transfers_include:
                cost = Decimal(0)
            else:
                return
        else:
            if self.holdings[t.asset].quantity:
                cost = self.holdings[t.asset].cost * (t.sell_quantity /
                                                      self.holdings[t.asset].quantity)
            else:
                cost = Decimal(0)

        self.holdings[t.asset].subtract_tokens(t.sell_quantity, cost)

        if t.is_disposal():
            tax_event = TaxCapitalGains(self.DISPOSAL_SECTION_104,
                                        None, t, cost.quantize(self.PRECISION))
            self.tax_events[self._which_tax_year(tax_event.date)].append(tax_event)

    def process_income(self):
        for t in self.transactions:
            if t.t_type in self.INCOME_TYPES:
                tax_event = TaxIncome(t)
                self.tax_events[self._which_tax_year(tax_event.date)].append(tax_event)

    def report_capital_gains(self, tax_year):
        log.info("==TAX SUMMARY %s/%s==", tax_year - 1, tax_year)
        log.info("--CAPITAL GAINS--")
        allowance = Decimal(self.ANNUAL_ALLOWANCE[tax_year])
        disposals = total_cost = total_proceeds = total_gain = Decimal(0)

        log.info("%s %s %s %s %s %s %s",
                 "Asset".ljust(7),
                 "Date".ljust(10),
                 "Disposal Type".ljust(28),
                 "Quantity".rjust(25),
                 "Cost".rjust(13),
                 "Proceeds".rjust(13),
                 "Gain".rjust(13))

        if tax_year in self.tax_events:
            for t in sorted(self.tax_events[tax_year]):
                if isinstance(t, TaxCapitalGains):
                    log.info(t)
                    disposals += 1
                    total_cost += t.cost
                    total_proceeds += t.proceeds
                    total_gain += t.gain

        log.info("Number of disposals=%s", disposals)
        log.info("Disposal proceeds=%s%s", config.sym(), '{:0,.2f}'.format(total_proceeds))
        if total_proceeds >= allowance * 4:
            log.warning("Assets sold are more than 4 times the annual allowance (%s%s), "
                        "this needs to be reported to HMRC",
                        config.sym(), '{:0,.2f}'.format(allowance * 4))
        log.info("Allowable costs=%s%s", config.sym(), '{:0,.2f}'.format(total_cost))

        gain = loss = Decimal(0)
        if total_gain >= 0:
            gain = total_gain
        else:
            loss = abs(total_gain)

        log.info("Gains in the year=%s%s", config.sym(), '{:0,.2f}'.format(gain))
        log.info("Losses in the year=%s%s", config.sym(), '{:0,.2f}'.format(loss))

        taxable_gain = Decimal(0)
        total_cg_tax = Decimal(0)
        if total_gain > allowance:
            taxable_gain = total_gain - allowance
            total_cg_tax = taxable_gain * self.CGT_RATE / 100

        log.info("--TAX ESTIMATE--")
        log.info("Taxable Gain=%s%s (-%s%s tax-free allowance)",
                 config.sym(), '{:0,.2f}'.format(taxable_gain),
                 config.sym(), '{:0,.2f}'.format(allowance))
        log.info("Capital Gains Tax=%s%s (%s%%)",
                 config.sym(), '{:0,.2f}'.format(total_cg_tax),
                 self.CGT_RATE)

    def report_income(self, tax_year):
        log.info("--INCOME--")

        total_income = Decimal(0)

        log.info("%s %s %s %s %s",
                 "Asset".ljust(7),
                 "Date".ljust(10),
                 "Income Type".ljust(28),
                 "Quantity".rjust(25),
                 "Amount".rjust(13))

        if tax_year in self.tax_events:
            for t in sorted(self.tax_events[tax_year]):
                if isinstance(t, TaxIncome):
                    log.info(t)
                    total_income += t.amount

        log.info("Total income=%s%s", config.sym(), '{:0,.2f}'.format(total_income))

    def report_holdings(self, value_asset):
        log.info("==CURRENT HOLDINGS==")

        total_cost = total_value = Decimal(0)

        log.info("%s %s %s %s  %s",
                 "Asset".ljust(7),
                 "Quantity".rjust(25),
                 "Cost".rjust(13),
                 "Value".rjust(13),
                 "Data Source")

        for h in sorted(self.holdings):
            if self.holdings[h].quantity > 0 or config.show_empty_wallets:
                cost = self.holdings[h].cost
                value, name, data_source = value_asset.get_current_value(self.holdings[h].asset,
                                                                         self.holdings[h].quantity)
                cost = cost.quantize(self.PRECISION)
                value = value.quantize(self.PRECISION)

                if data_source:
                    log.info("%s %s %s %s  %s (%s)",
                             self.holdings[h].asset.ljust(7),
                             self.holdings[h].format_quantity().rjust(25),
                             (config.sym() + '{:0,.2f}'.format(cost)).rjust(13),
                             (config.sym() + '{:0,.2f}'.format(value)).rjust(13),
                             data_source,
                             name)
                else:
                    log.info("%s %s %s %s  -",
                             self.holdings[h].asset.ljust(7),
                             self.holdings[h].format_quantity().rjust(25),
                             (config.sym() + '{:0,.2f}'.format(cost)).rjust(13),
                             (config.sym() + '{:0,.2f}'.format(value)).rjust(13))

                total_cost += cost
                total_value += value

        log.info("Total cost=%s%s", config.sym(), '{:0,.2f}'.format(total_cost))
        log.info("Total value=%s%s", config.sym(), '{:0,.2f}'.format(total_value))

    def _which_tax_year(self, timestamp):
        if timestamp > datetime(timestamp.year, 4, 5, tzinfo=config.TZ_LOCAL):
            tax_year = timestamp.year + 1
        else:
            tax_year = timestamp.year

        if tax_year not in self.tax_events:
            self.tax_events[tax_year] = []

        return tax_year

class TaxEvent(object):
    def __init__(self, date, asset):
        self.date = date
        self.asset = asset

    def __eq__(self, other):
        return (self.asset, self.date) == (other.asset, other.date)

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return (self.asset, self.date) < (other.asset, other.date)

class TaxCapitalGains(TaxEvent):
    def __init__(self, disposal_type, b, s, cost):
        super(TaxCapitalGains, self).__init__(s.timestamp, s.asset)
        self.disposal_type = disposal_type
        self.quantity = s.sell_quantity
        self.cost = cost
        self.proceeds = s.proceeds
        self.gain = self.proceeds - self.cost
        self.acquisition_date = b.timestamp if b else None

        log.debug(" Gain=%s%s (%s%s - %s%s)",
                  config.sym(), '{:0,.2f}'.format(self.gain),
                  config.sym(), '{:0,.2f}'.format(self.proceeds),
                  config.sym(), '{:0,.2f}'.format(self.cost))

    @staticmethod
    def format_disposal(disposal_type, date):
        if disposal_type == TaxCalculator.DISPOSAL_BED_AND_BREAKFAST:
            return disposal_type + " (" + date.strftime('%d/%m/%Y') + ")"

        return disposal_type

    def __str__(self):
        return self.asset.ljust(7) + " " + \
               self.date.strftime('%d/%m/%Y') + " " + \
               self.format_disposal(self.disposal_type, self.acquisition_date).ljust(28) + " " + \
               '{:0,f}'.format(self.quantity.normalize()).rjust(25) + " " + \
               (config.sym() + '{:0,.2f}'.format(self.cost)).rjust(13) + " " + \
               (config.sym() + '{:0,.2f}'.format(self.proceeds)).rjust(13) + " " + \
               (config.sym() + '{:0,.2f}'.format(self.gain)).rjust(13)

class TaxIncome(TaxEvent):
    def __init__(self, b):
        super(TaxIncome, self).__init__(b.timestamp, b.asset)
        self.type = b.t_type
        self.quantity = b.buy_quantity
        self.amount = b.cost

    def __str__(self):
        return self.asset.ljust(7) + " " + \
               self.date.strftime('%d/%m/%Y') + " " + \
               self.type.ljust(28) + " " + \
               '{:0,f}'.format(self.quantity.normalize()).rjust(25) + " " + \
               (config.sym() + '{:0,.2f}'.format(self.amount)).rjust(13)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Coinfloor"

def parse_coinfloor_trades(in_row):
    if in_row[7] == "Buy":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[3],
                                    buy_asset=in_row[1],
                                    sell_quantity=in_row[5],
                                    sell_asset=in_row[2],
                                    fee_quantity=in_row[6],
                                    fee_asset=in_row[2],
                                    wallet=WALLET)
    elif in_row[7] == "Sell":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[5],
                                    buy_asset=in_row[2],
                                    sell_quantity=in_row[3],
                                    sell_asset=in_row[1],
                                    fee_quantity=in_row[6],
                                    fee_asset=in_row[2],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised Order Type: " + in_row[7])

def parse_coinfloor_deposits_withdrawals(in_row):
    if in_row[3] == "Deposit":
        return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[1],
                                    buy_asset=in_row[2],
                                    wallet=WALLET)
    elif in_row[3] == "Withdrawal":
        return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                    DataParser.parse_timestamp(in_row[0]),
                                    sell_quantity=in_row[1],
                                    sell_asset=in_row[2],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised Type: " + in_row[3])

DataParser(DataParser.TYPE_EXCHANGE,
           "Coinfloor Trades",
           ['Date & Time', 'Base Asset', 'Counter Asset', 'Amount', 'Price', 'Total', 'Fee',
            'Order Type'],
           row_handler=parse_coinfloor_trades)

DataParser(DataParser.TYPE_EXCHANGE,
           "Coinfloor Deposits/Withdrawals",
           ['Date & Time', 'Amount', 'Asset', 'Type'],
           row_handler=parse_coinfloor_deposits_withdrawals)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import copy
from decimal import Decimal

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Gatehub"

log = logging.getLogger()

def parse_gatehub(all_in_row_orig):
    all_in_row = copy.deepcopy(all_in_row_orig)
    t_records = []

    for i, in_row in enumerate(all_in_row):
        t_type = ""
        buy_quantity = None
        buy_asset = ""
        sell_quantity = None
        sell_asset = ""
        fee_quantity = None
        fee_asset = ""

        if not in_row[1] or not in_row[3]:
            # Skip if already matched, or Amount missing
            t_records.append(None)
            continue

        if in_row[2] == "payment":
            if Decimal(in_row[3]) < 0:
                t_type = TransactionOutRecord.TYPE_WITHDRAWAL
                sell_quantity = abs(Decimal(in_row[3]))
                sell_asset = in_row[4]
            else:
                t_type = TransactionOutRecord.TYPE_DEPOSIT
                buy_quantity = in_row[3]
                buy_asset = in_row[4]

            fee_quantity, fee_asset = find_same_tx(all_in_row, i,
                                                   "ripple_network_fee",
                                                   [sell_asset, buy_asset])
        elif in_row[2] == "exchange":
            t_type = TransactionOutRecord.TYPE_TRADE
            if Decimal(in_row[3]) < 0:
                sell_quantity = abs(Decimal(in_row[3]))
                sell_asset = in_row[4]

                buy_quantity, buy_asset = find_same_tx(all_in_row, i,
                                                       "exchange")
            else:
                buy_quantity = in_row[3]
                buy_asset = in_row[4]

                sell_quantity, sell_asset = find_same_tx(all_in_row, i,
                                                         "exchange")

            if sell_quantity is None or buy_quantity is None:
                # Skip if buy or sell is missing from Trade
                continue

            fee_quantity, fee_asset = find_same_tx(all_in_row, i,
                                                   "ripple_network_fee",
                                                   [sell_asset, buy_asset])
        elif in_row[2] == "ripple_network_fee":
            # Fees which are not associated with a payment or exchange are added
            # as a Spend
            t_type = TransactionOutRecord.TYPE_SPEND
            sell_quantity = abs(Decimal(in_row[3]))
            sell_asset = in_row[4]
        else:
            raise ValueError("Unrecognised Type: " + in_row[2])

        t_record = TransactionOutRecord(t_type,
                                        DataParser.parse_timestamp(in_row[0]),
                                        buy_quantity=buy_quantity,
                                        buy_asset=buy_asset,
                                        sell_quantity=sell_quantity,
                                        sell_asset=sell_asset,
                                        fee_quantity=fee_quantity,
                                        fee_asset=fee_asset,
                                        wallet=WALLET)
        t_records.append(t_record)
        in_row[1] = None    # Remove hash to prevent future matches

    for i, in_row in enumerate(all_in_row):
        # Check for unmatched transactions
        if in_row[1]:
            log.warning("Skipping row[%s]: %s", i + 2, in_row)

    return t_records

def find_same_tx(all_in_row, i, tx_type, currencies=None):
    tx_hash = all_in_row[i][1]
    quantity = None
    asset = ""

    for _, in_row in enumerate(all_in_row[i + 1:len(all_in_row)]):
        if in_row[1] == tx_hash and in_row[2] == tx_type and \
                (tx_type == "ripple_network_fee" and in_row[4] in currencies or
                 tx_type != "ripple_network_fee"):
            quantity = abs(Decimal(in_row[3]))
            asset = in_row[4]
            in_row[1] = None    # Remove hash to prevent future matches
            break

    return quantity, asset

DataParser(DataParser.TYPE_EXCHANGE,
           "GateHub (Ripple)",
           ['Time', 'TX hash', 'Type', 'Amount', 'Currency', 'Currency Issuer Address',
            'Currency Issuer Name', 'Balance'],
           all_handler=parse_gatehub)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Cryptopia"

def parse_cryptopia_deposits(in_row):
    return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                DataParser.parse_timestamp(in_row[7]),
                                buy_quantity=in_row[2],
                                buy_asset=in_row[1],
                                wallet=WALLET)

def parse_cryptopia_withdrawals(in_row):
    return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                DataParser.parse_timestamp(in_row[7]),
                                sell_quantity=Decimal(in_row[2]) - Decimal(in_row[3]),
                                sell_asset=in_row[1],
                                fee_quantity=in_row[3],
                                fee_asset=in_row[1],
                                wallet=WALLET)

def parse_cryptopia_trades(in_row):
    if in_row[2] == "Buy":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[7]),
                                    buy_quantity=in_row[4],
                                    buy_asset=in_row[1].split('/')[0],
                                    sell_quantity=Decimal(in_row[3]) * Decimal(in_row[4]),
                                    sell_asset=in_row[1].split('/')[1],
                                    fee_quantity=in_row[6],
                                    fee_asset=in_row[1].split('/')[0],
                                    wallet=WALLET)
    elif in_row[2] == "Sell":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[7]),
                                    buy_quantity=Decimal(in_row[3]) * Decimal(in_row[4]),
                                    buy_asset=in_row[1].split('/')[1],
                                    sell_quantity=in_row[4],
                                    sell_asset=in_row[1].split('/')[0],
                                    fee_quantity=in_row[6],
                                    fee_asset=in_row[1].split('/')[1],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised Type: " + in_row[2])

DataParser(DataParser.TYPE_EXCHANGE,
           "Cryptopia Deposits",
           ['#', 'Currency', 'Amount', 'Status', 'Type', 'Transaction', 'Conf.', 'Timestamp'],
           row_handler=parse_cryptopia_deposits)

DataParser(DataParser.TYPE_EXCHANGE,
           "Cryptopia Withdrawals",
           ['#', 'Currency', 'Amount', 'Fee', 'Status', 'TransactionId', 'Address', 'Timestamp'],
           row_handler=parse_cryptopia_withdrawals)

DataParser(DataParser.TYPE_EXCHANGE,
           "Cryptopia Trades",
           ['#', 'Market', 'Type', 'Rate', 'Amount', 'Total', 'Fee', 'Timestamp'],
           row_handler=parse_cryptopia_trades)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import sys
import csv
import copy
from decimal import Decimal

import dateutil.parser

from .config import config
from .record import TransactionInRecord

PRECISION = Decimal('0.00')

log = logging.getLogger()

def load_transaction_records(import_file):
    log.info("==IMPORT TRANSACTION RECORDS==")
    transaction_records = []

    if sys.version_info[0] < 3:
        # Special handling required for utf-8 encoded csv files
        reader = csv.reader(_utf_8_encoder(import_file))
    else:
        reader = csv.reader(import_file)

    next(reader, None) # skip headers
    for row in reader:
        record = _parse_row(row)
        record.validate()
        record.normalise_to_localtime()
        record.include_fees()
        log.debug(record)
        transaction_records.append(record)

    log.info("Total transaction records=%s", len(transaction_records))
    return transaction_records

def _utf_8_encoder(unicode_csv_data):
    for line in unicode_csv_data:
        yield line.encode('utf-8')

def _parse_row(row):
    timestamp = dateutil.parser.parse(row[11], tzinfos=config.TZ_INFOS)

    if timestamp.tzinfo is None:
        # Default to UTC if no timezone is specified
        timestamp = timestamp.replace(tzinfo=config.TZ_UTC)

    return TransactionInRecord(row[0],
                               timestamp,
                               buy_quantity=_strip_non_digits(row[1]) if row[1] else None,
                               buy_asset=row[2],
                               buy_value=_strip_non_digits(row[3]) if row[3] else None,
                               sell_quantity=_strip_non_digits(row[4]) if row[4] else None,
                               sell_asset=row[5],
                               sell_value=_strip_non_digits(row[6]) if row[6] else None,
                               fee_quantity=_strip_non_digits(row[7]) if row[7] else None,
                               fee_asset=row[8],
                               fee_value=_strip_non_digits(row[9]) if row[9] else None,
                               wallet=row[10])

def _strip_non_digits(string):
    return string.strip('£€$').replace(',', '')

def _format_quantity(quantity):
    if quantity is None:
        return '-'

    return '{:0,f}'.format(quantity.normalize())

def _format_value(value):
    if value is not None:
        return ' (' + config.sym() + '{:0,.2f} {})'.format(value, config.CCY)

    return ''

def _format_pooled(pooled):
    if len(pooled) > 1:
        return " [" + str(len(pooled)) + "]"

    return ''

class TransactionHistory(object):
    def __init__(self, transaction_records):
        self.transaction_records = transaction_records

    def split_transaction_records(self, value_asset):
        log.debug("==SPLIT TRANSACTION RECORDS==")

        transactions = []
        for tr in self.transaction_records:
            if tr.sell_asset and tr.sell_asset not in config.fiat_list:
                value = self.which_asset_value(tr, value_asset)
                sell = Sell(tr.line_num, tr.t_type, tr.wallet, tr.timestamp,
                            tr.sell_quantity, tr.sell_asset, value)
                transactions.append(sell)
                log.debug(sell)

            if tr.buy_asset and tr.buy_asset not in config.fiat_list:
                value = self.which_asset_value(tr, value_asset)
                buy = Buy(tr.line_num, tr.t_type, tr.wallet, tr.timestamp,
                          tr.buy_quantity, tr.buy_asset, value)
                transactions.append(buy)
                log.debug(buy)

        log.debug("Total transactions=%s", len(transactions))
        return transactions

    @staticmethod
    def which_asset_value(tr, value_asset):
        value = None
        if tr.t_type in tr.BUY_TYPES:
            value = value_asset.get_value(tr.buy_asset,
                                          tr.timestamp,
                                          tr.buy_quantity,
                                          tr.buy_value)
        elif tr.t_type in tr.SELL_TYPES:
            value = value_asset.get_value(tr.sell_asset,
                                          tr.timestamp,
                                          tr.sell_quantity,
                                          tr.sell_value)
        elif tr.t_type == tr.TYPE_TRADE:
            if config.trade_asset_type == config.TRADE_ASSET_TYPE_BUY:
                value = value_asset.get_value(tr.buy_asset,
                                              tr.timestamp,
                                              tr.buy_quantity,
                                              tr.buy_value)
            elif config.trade_asset_type == config.TRADE_ASSET_TYPE_SELL:
                value = value_asset.get_value(tr.sell_asset,
                                              tr.timestamp,
                                              tr.sell_quantity,
                                              tr.sell_value)
            else:
                pos_sell_asset = pos_buy_asset = len(config.asset_priority) + 1

                if tr.sell_asset in config.asset_priority:
                    pos_sell_asset = config.asset_priority.index(tr.sell_asset)
                if tr.buy_asset in config.asset_priority:
                    pos_buy_asset = config.asset_priority.index(tr.buy_asset)

                if pos_sell_asset <= pos_buy_asset:
                    value = value_asset.get_value(tr.sell_asset,
                                                  tr.timestamp,
                                                  tr.sell_quantity,
                                                  tr.sell_value)
                else:
                    value = value_asset.get_value(tr.buy_asset,
                                                  tr.timestamp,
                                                  tr.buy_quantity,
                                                  tr.buy_value)
        if value is not None:
            value = value.quantize(PRECISION)

        return value

class TransactionBase(object):
    def __init__(self, tid, t_type, asset, wallet, timestamp):
        self.tid = [tid, 0]
        self.t_type = t_type
        self.asset = asset
        self.wallet = wallet
        self.timestamp = timestamp
        self.matched = False
        self.pooled = []

    def _format_match_status(self):
        return " (M)" if self.matched else ""

    def _format_tid(self):
        if self.tid[1] == 0:
            return str(self.tid[0])
        return str(self.tid[0]) + "." + str(self.tid[1])

    def __eq__(self, other):
        return (self.asset, self.timestamp) == (other.asset, other.timestamp)

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return (self.asset, self.timestamp) < (other.asset, other.timestamp)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result

class Buy(TransactionBase):
    TYPE_DEPOSIT = TransactionInRecord.TYPE_DEPOSIT
    TYPE_MINING = TransactionInRecord.TYPE_MINING
    TYPE_INCOME = TransactionInRecord.TYPE_INCOME
    TYPE_GIFT_RECEIVED = TransactionInRecord.TYPE_GIFT_RECEIVED
    TYPE_TRADE = TransactionInRecord.TYPE_TRADE

    ACQUISITION_TYPES = {TYPE_MINING, TYPE_INCOME, TYPE_GIFT_RECEIVED, TYPE_TRADE}

    def __init__(self, tid, t_type, wallet, timestamp, buy_quantity, buy_asset, buy_value):
        super(Buy, self).__init__(tid, t_type, buy_asset, wallet, timestamp)
        self.buy_quantity = buy_quantity
        self.cost = buy_value

        self.pooled.append(copy.deepcopy(self))

    def __iadd__(self, other):
        # Pool buys
        if self.asset != other.asset:
            raise Exception
        self.buy_quantity += other.buy_quantity
        self.cost += other.cost

        if other.timestamp < self.timestamp:
            # Keep timestamp of earliest transaction
            self.timestamp = other.timestamp

        if other.wallet != self.wallet:
            self.wallet = "<pooled>"

        self.pooled.append(other)
        return self

    def is_acquisition(self):
        return self.t_type in self.ACQUISITION_TYPES

    def split_buy(self, sell_quantity):
        remainder = copy.deepcopy(self)
        remainder.tid[1] += 1

        self.cost = Decimal(self.cost * (sell_quantity /
                                         self.buy_quantity)).quantize(PRECISION)
        self.buy_quantity = sell_quantity

        remainder.cost = remainder.cost - self.cost
        remainder.buy_quantity = remainder.buy_quantity - sell_quantity

        log.debug("split: %s", self)
        log.debug("split: %s", remainder)

        return remainder

    def __str__(self):
        return "[TID:" + self._format_tid() + "] " + \
               type(self).__name__ + " (" + \
               self.t_type + "): " + \
               _format_quantity(self.buy_quantity) + " " + \
               self.asset + \
               _format_value(self.cost) + " \"" + \
               self.wallet + "\" " + \
               self.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z') + \
               _format_pooled(self.pooled) + \
               self._format_match_status()

class Sell(TransactionBase):
    TYPE_WITHDRAWAL = TransactionInRecord.TYPE_WITHDRAWAL
    TYPE_SPEND = TransactionInRecord.TYPE_SPEND
    TYPE_GIFT_SENT = TransactionInRecord.TYPE_GIFT_SENT
    TYPE_CHARITY_SENT = TransactionInRecord.TYPE_CHARITY_SENT
    TYPE_TRADE = TransactionInRecord.TYPE_TRADE

    DISPOSAL_TYPES = {TYPE_SPEND, TYPE_GIFT_SENT, TYPE_CHARITY_SENT, TYPE_TRADE}

    def __init__(self, tid, t_type, wallet, timestamp, sell_quantity, sell_asset, sell_value):
        super(Sell, self).__init__(tid, t_type, sell_asset, wallet, timestamp)
        self.sell_quantity = sell_quantity
        self.proceeds = sell_value

        self.pooled.append(copy.deepcopy(self))

    def __iadd__(self, other):
	# Pool sells
        if self.asset != other.asset:
            raise Exception
        self.sell_quantity += other.sell_quantity
        self.proceeds += other.proceeds

        if other.timestamp > self.timestamp:
            # Keep timestamp of latest transaction
            self.timestamp = other.timestamp

        if other.wallet != self.wallet:
            self.wallet = "<pooled>"

        self.pooled.append(other)
        return self

    def is_disposal(self):
        return self.t_type in self.DISPOSAL_TYPES

    def split_sell(self, buy_quantity):
        remainder = copy.deepcopy(self)
        remainder.tid[1] += 1

        self.proceeds = Decimal(self.proceeds * (buy_quantity /
                                                 self.sell_quantity)).quantize(PRECISION)
        self.sell_quantity = buy_quantity

        remainder.proceeds = remainder.proceeds - self.proceeds
        remainder.sell_quantity = remainder.sell_quantity - buy_quantity

        log.debug("split: %s", self)
        log.debug("split: %s", remainder)

        return remainder

    def __str__(self):
        return "[TID:" + self._format_tid() + "] " + \
               type(self).__name__ + " (" + \
               self.t_type + "): " + \
               _format_quantity(self.sell_quantity) + " " + \
               self.asset + \
               _format_value(self.proceeds) + " \"" + \
               self.wallet + "\" " + \
               self.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z') + \
               _format_pooled(self.pooled) + \
               self._format_match_status()



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import argparse
import csv
import sys
import io

import xlrd

from ..version import __version__
from ..config import config
from .dataparser import DataParser
from .parsers import *
from .out_record import TransactionOutRecord

CSV_DELIMITERS = (',', ';')

logging.basicConfig(stream=sys.stderr,
                    level=logging.INFO,
                    format='[%(asctime)s.%(msecs)03d] %(levelname)s -- : %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')
log = logging.getLogger()

def _open_excel_file(workbook):
    sheet = workbook.sheet_by_index(0)

    reader = _get_cell_values(sheet.get_rows(), workbook)
    all_in_row, t_records, in_header = _parse_file(reader)
    TransactionOutRecord.csv_file(all_in_row, t_records, in_header)
    workbook.release_resources()
    del workbook

def _get_cell_values(rows, workbook):
    for row in rows:
        yield [_convert_cell(cell, workbook) for cell in row]

def _convert_cell(cell, workbook):
    if cell.ctype == xlrd.XL_CELL_DATE:
        value = xlrd.xldate.xldate_as_datetime(cell.value, workbook.datemode). \
                    strftime('%Y-%m-%dT%H:%M:%S %Z')
    else:
        value = str(cell.value)

    return value

def _open_csv_file(filename, delimiter):
    with io.open(filename, newline='', encoding='utf-8-sig') as data_file:
        if sys.version_info[0] < 3:
            # special handling required for utf-8 encoded csv files
            reader = csv.reader(_utf_8_encoder(data_file), delimiter=delimiter)
        else:
            reader = csv.reader(data_file, delimiter=delimiter)

        all_in_row, t_records, in_header = _parse_file(reader)
        TransactionOutRecord.csv_file(all_in_row, t_records, in_header)

    data_file.close()

def _utf_8_encoder(unicode_csv_data):
    for line in unicode_csv_data:
        yield line.encode('utf-8')

def _parse_file(reader):
    parser = None
    # header might not be on first line
    for row in range(5):
        try:
            log.debug("Row:%d", row)
            parser = DataParser.match_header(next(reader))
        except KeyError:
            continue
        except StopIteration:
            pass
        else:
            break

    if parser is None:
        raise KeyError("Data file format unrecognised")

    if parser.row_handler:
        all_in_row = []
        t_records = []

        for in_row in reader:
            all_in_row.append(in_row)
            t_records.append(parser.row_handler(in_row, *parser.args))
    else:
        # all rows handled together
        all_in_row = list(reader)
        t_records = parser.all_handler(all_in_row, *parser.args)

    return all_in_row, t_records, parser.in_header

def main():
    parser = argparse.ArgumentParser(epilog="supported data file formats:\n" + \
                                     DataParser.format_parsers(),
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("filename",
                        type=str,
                        nargs='+',
                        help="filename of data file")
    parser.add_argument("-v",
                        "--version",
                        action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument("-d",
                        "--debug",
                        action='store_true',
                        help="enabled debug logging")
    parser.add_argument("-uc",
                        "--unconfirmed",
                        action='store_true',
                        help="include unconfirmed transactions")
    parser.add_argument("-ca",
                        dest="cryptoasset",
                        type=str,
                        help="specify a cryptoasset symbol, if it cannot be identified "
                             "automatically")
    parser.add_argument("-nh",
                        "--noheader",
                        action='store_true',
                        help="exclude header from CSV output")
    parser.add_argument("-a",
                        "--append",
                        action='store_true',
                        help="append output as new columns to the input data file")
    parser.add_argument("--format",
                        choices=['CSV', 'RECAP'],
                        default='CSV',
                        type=str.upper,
                        help="specify the output format")
    parser.add_argument("-s",
                        "--sort",
                        action='store_true',
                        help="sort output by timestamp")

    config.args = parser.parse_args()

    if config.args.debug:
        log.setLevel(logging.DEBUG)
        config.output_config(parser.prog)

    for filename in config.args.filename:
        try:
            log.debug("EXCEL")
            workbook = xlrd.open_workbook(filename)
        except xlrd.XLRDError:
            for delimiter in CSV_DELIMITERS:
                log.debug("CSV, delimiter='%s'", delimiter)
                try:
                    key_error = None
                    _open_csv_file(filename, delimiter=delimiter)
                except KeyError as key_error:
                    # Try with next delimiter
                    continue
                else:
                    break

            if key_error is not None:
                raise
        else:
            _open_excel_file(workbook)

        config.args.noheader = True



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "TradeSatoshi"

def parse_tradesatoshi_deposits2(in_row):
    return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                DataParser.parse_timestamp(in_row[0]),
                                buy_quantity=in_row[3],
                                buy_asset=in_row[2],
                                wallet=WALLET)

def parse_tradesatoshi_deposits(in_row):
    return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                DataParser.parse_timestamp(in_row[7]),
                                buy_quantity=in_row[3],
                                buy_asset=in_row[2],
                                wallet=WALLET)

def parse_tradesatoshi_withdrawals2(in_row):
    return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                DataParser.parse_timestamp(in_row[0]),
                                sell_quantity=Decimal(in_row[3]),
                                sell_asset=in_row[2],
                                wallet=WALLET)

def parse_tradesatoshi_withdrawals(in_row):
    return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                DataParser.parse_timestamp(in_row[10]),
                                sell_quantity=Decimal(in_row[3]) - Decimal(in_row[4]),
                                sell_asset=in_row[2],
                                fee_quantity=in_row[4],
                                fee_asset=in_row[2],
                                wallet=WALLET)

def parse_tradesatoshi_trades(in_row, *_):
    if in_row[2] == "Buy":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[6]),
                                    buy_quantity=in_row[3],
                                    buy_asset=in_row[1].split('/')[0],
                                    sell_quantity=Decimal(in_row[3]) * Decimal(in_row[4]),
                                    sell_asset=in_row[1].split('/')[1],
                                    fee_quantity=in_row[5],
                                    fee_asset=in_row[1].split('/')[1],
                                    wallet=WALLET)
    elif in_row[2] == "Sell":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[6]),
                                    buy_quantity=Decimal(in_row[3]) * Decimal(in_row[4]),
                                    buy_asset=in_row[1].split('/')[1],
                                    sell_quantity=in_row[3],
                                    sell_asset=in_row[1].split('/')[0],
                                    fee_quantity=in_row[5],
                                    fee_asset=in_row[1].split('/')[1],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised TradeType: " + in_row[2])

DataParser(DataParser.TYPE_EXCHANGE,
           "TradeSatoshi Deposits",
           ['TimeStamp', 'Currency', 'Symbol', 'Amount', 'Confirmation', 'TxId'],
           row_handler=parse_tradesatoshi_deposits2)

DataParser(DataParser.TYPE_EXCHANGE,
           "TradeSatoshi Deposits",
           ['Id', 'Currency', 'Symbol', 'Amount', 'Status', 'Confirmations', 'TxId', 'TimeStamp'],
           row_handler=parse_tradesatoshi_deposits)

DataParser(DataParser.TYPE_EXCHANGE,
           "TradeSatoshi Withdrawals",
           ['TimeStamp', 'Currency', 'Symbol', 'Amount', 'Confirmation', 'TxId', 'Address',
            'PaymentId', 'Status'],
           row_handler=parse_tradesatoshi_withdrawals2)

DataParser(DataParser.TYPE_EXCHANGE,
           "TradeSatoshi Withdrawals",
           ['Id', 'User', 'Symbol', 'Amount', 'Fee', 'Net Amount', 'Status', 'Confirmations',
            'TxId', 'Address', 'TimeStamp'],
           row_handler=parse_tradesatoshi_withdrawals)

DataParser(DataParser.TYPE_EXCHANGE,
           "TradeSatoshi Trades",
           ['Id', 'TradePair', lambda c: c in ('TradeType', 'TradeHistoryType'), 'Amount', 'Rate',
            'Fee', lambda c: c.lower() == 'timestamp', 'IsApi'],
           row_handler=parse_tradesatoshi_trades)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
import csv
import sys

from ..record import TransactionRecordBase
from ..config import config

log = logging.getLogger()

EXCEL_PRECISION = 15

class TransactionOutRecord(TransactionRecordBase):
    RECAP_TYPE_MAPPING = {TransactionRecordBase.TYPE_DEPOSIT: 'Deposit',
                          TransactionRecordBase.TYPE_MINING: 'Mining',
                          TransactionRecordBase.TYPE_INCOME: 'Income',
                          TransactionRecordBase.TYPE_GIFT_RECEIVED: 'Gift',
                          TransactionRecordBase.TYPE_WITHDRAWAL: 'Withdrawal',
                          TransactionRecordBase.TYPE_SPEND: 'Purchase',
                          TransactionRecordBase.TYPE_GIFT_SENT: 'Gift',
                          TransactionRecordBase.TYPE_CHARITY_SENT: 'Donation',
                          TransactionRecordBase.TYPE_TRADE: 'Trade'}

    BITTYTAX_OUT_HEADER = ['Type',
                           'Buy Quantity', 'Buy Asset', 'Buy Value',
                           'Sell Quantity', 'Sell Asset', 'Sell Value',
                           'Fee Quantity', 'Fee Asset', 'Fee Value',
                           'Wallet', 'Timestamp']

    RECAP_OUT_HEADER = ['Type', 'Date',
                        'InOrBuyAmount', 'InOrBuyCurrency',
                        'OutOrSellAmount', 'OutOrSellCurrency',
                        'FeeAmount', 'FeeCurrency']

    def to_csv(self):
        if config.args.format == config.FORMAT_RECAP:
            return self._to_recap()
        else:
            return self._to_bittytax()

    def _to_bittytax(self):
        if self.buy_quantity is not None and \
                len(self.buy_quantity.normalize().as_tuple().digits) > EXCEL_PRECISION:
            log.warning("%d-digit precision exceeded! %s", EXCEL_PRECISION, self)

        if self.sell_quantity is not None and \
                len(self.sell_quantity.normalize().as_tuple().digits) > EXCEL_PRECISION:
            log.warning("%d-digit precision exceeded! %s", EXCEL_PRECISION, self)

        if self.fee_quantity is not None and \
                len(self.fee_quantity.normalize().as_tuple().digits) > EXCEL_PRECISION:
            log.warning("%d-digit precision exceeded! %s", EXCEL_PRECISION, self)

        return [self.t_type,
                '{0:f}'.format(self.buy_quantity) if self.buy_quantity is not None else None,
                self.buy_asset,
                '{0:f}'.format(self.buy_value) if self.buy_value is not None else None,
                '{0:f}'.format(self.sell_quantity) if self.sell_quantity is not None else None,
                self.sell_asset,
                '{0:f}'.format(self.sell_value) if self.sell_value is not None else None,
                '{0:f}'.format(self.fee_quantity) if self.fee_quantity is not None else None,
                self.fee_asset,
                '{0:f}'.format(self.fee_value) if self.fee_value is not None else None,
                self.wallet,
                self.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z')]

    def _to_recap(self):
        return [self.RECAP_TYPE_MAPPING[self.t_type],
                self.timestamp.strftime('%Y-%m-%d %H:%M:%S'),
                '{0:f}'.format(self.buy_quantity) if self.buy_quantity is not None else None,
                self.buy_asset,
                '{0:f}'.format(self.sell_quantity) if self.sell_quantity is not None else None,
                self.sell_asset,
                '{0:f}'.format(self.fee_quantity) if self.fee_quantity is not None else None,
                self.fee_asset]

    @classmethod
    def out_header(cls):
        if config.args.format == config.FORMAT_RECAP:
            return cls.RECAP_OUT_HEADER
        else:
            return cls.BITTYTAX_OUT_HEADER

    @classmethod
    def in_header(cls, in_header):
        if config.args.format == config.FORMAT_RECAP:
            return [name if name not in cls.out_header() else name + '_' for name in in_header]
        else:
            return in_header

    @classmethod
    def prepend(cls):
        if config.args.format == config.FORMAT_RECAP:
            return True
        else:
            return False

    @classmethod
    def sort_key(cls):
        if config.args.format == config.FORMAT_RECAP:
            return lambda c: c[cls.RECAP_OUT_HEADER.index('Date')]
        else:
            return lambda c: c[-1]

    @classmethod
    def csv_file(cls, all_in_row, t_records, in_header):
        if config.args.append:
            if cls.prepend():
                out_rows = [t_record.to_csv() + in_row if t_record
                            else [None] * len(cls.out_header()) + in_row
                            for in_row, t_record in zip(all_in_row, t_records)]
            else:
                out_rows = [in_row + t_record.to_csv() if t_record else in_row
                            for in_row, t_record in zip(all_in_row, t_records)]
        else:
            out_rows = [t_record.to_csv() for t_record in t_records if t_record]

        if config.args.sort:
            out_rows.sort(key=cls.sort_key(), reverse=False)

        writer = csv.writer(sys.stdout, lineterminator='\n')
        if not config.args.noheader:
            if config.args.append:
                if cls.prepend():
                    writer.writerow(cls.out_header() + cls.in_header(in_header))
                else:
                    writer.writerow(cls.in_header(in_header) + cls.out_header())
            else:
                writer.writerow(cls.out_header())

        writer.writerows(out_rows)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Bitstamp"

def parse_bitstamp(in_row):
    if in_row[0] in ("Ripple deposit", "Deposit"):
        return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                    DataParser.parse_timestamp(in_row[1]),
                                    buy_quantity=in_row[3].split(' ')[0],
                                    buy_asset=in_row[3].split(' ')[1],
                                    wallet=WALLET)
    elif in_row[0] in ("Ripple payment", "Withdrawal"):
        return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                    DataParser.parse_timestamp(in_row[1]),
                                    sell_quantity=in_row[3].split(' ')[0],
                                    sell_asset=in_row[3].split(' ')[1],
                                    wallet=WALLET)
    elif in_row[0] == "Market":
        if in_row[7] == "Buy":
            return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                        DataParser.parse_timestamp(in_row[1]),
                                        buy_quantity=in_row[3].split(' ')[0],
                                        buy_asset=in_row[3].split(' ')[1],
                                        sell_quantity=in_row[4].split(' ')[0],
                                        sell_asset=in_row[4].split(' ')[1],
                                        fee_quantity=in_row[6].split(' ')[0],
                                        fee_asset=in_row[6].split(' ')[1],
                                        wallet=WALLET)
        elif in_row[7] == "Sell":
            return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                        DataParser.parse_timestamp(in_row[1]),
                                        buy_quantity=in_row[4].split(' ')[0],
                                        buy_asset=in_row[4].split(' ')[1],
                                        sell_quantity=in_row[3].split(' ')[0],
                                        sell_asset=in_row[3].split(' ')[1],
                                        fee_quantity=in_row[6].split(' ')[0],
                                        fee_asset=in_row[6].split(' ')[1],
                                        wallet=WALLET)
        else:
            raise ValueError("Unrecognised Sub Type: " + in_row[7])
    else:
        raise ValueError("Unrecognised Type: " + in_row[0])

DataParser(DataParser.TYPE_EXCHANGE,
           "Bitstamp",
           ['Type', 'Datetime', 'Account', 'Amount', 'Value', 'Rate', 'Fee', 'Sub Type'],
           row_handler=parse_bitstamp)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal

from .config import config

class TransactionRecordBase(object):
    TYPE_DEPOSIT = 'Deposit'
    TYPE_MINING = 'Mining'
    TYPE_INCOME = 'Income'
    TYPE_GIFT_RECEIVED = 'Gift-Received'
    TYPE_WITHDRAWAL = 'Withdrawal'
    TYPE_SPEND = 'Spend'
    TYPE_GIFT_SENT = 'Gift-Sent'
    TYPE_CHARITY_SENT = 'Charity-Sent'
    TYPE_TRADE = 'Trade'

    BUY_TYPES = (TYPE_MINING, TYPE_INCOME, TYPE_GIFT_RECEIVED)
    SELL_TYPES = (TYPE_SPEND, TYPE_GIFT_SENT, TYPE_CHARITY_SENT)

    cnt = 0

    def __init__(self, t_type, timestamp,
                 buy_quantity=None, buy_asset="", buy_value=None,
                 sell_quantity=None, sell_asset="", sell_value=None,
                 fee_quantity=None, fee_asset="", fee_value=None,
                 wallet=""):

        TransactionRecordBase.cnt += 1
        self.line_num = TransactionRecordBase.cnt
        self.t_type = t_type
        self.buy_quantity = Decimal(buy_quantity) if buy_quantity is not None else None
        self.buy_asset = buy_asset
        self.buy_value = Decimal(buy_value) if buy_value is not None else None
        self.sell_quantity = Decimal(sell_quantity) if sell_quantity is not None else None
        self.sell_asset = sell_asset
        self.sell_value = Decimal(sell_value) if sell_value is not None else None
        self.fee_quantity = Decimal(fee_quantity) if fee_quantity is not None else None
        self.fee_asset = fee_asset
        self.fee_value = Decimal(fee_value) if fee_value is not None else None
        self.wallet = wallet
        self.timestamp = timestamp

    @staticmethod
    def format_quantity(quantity):
        if quantity is None:
            return '-'

        return '{:0,f}'.format(quantity.normalize())

    @staticmethod
    def format_value(value):
        if value is not None:
            return ' (' + config.sym() + '{:0,.2f} {})'.format(value, config.CCY)

        return ''

    def __str__(self):
        if self.t_type in self.BUY_TYPES or self.t_type == self.TYPE_DEPOSIT:
            return "[Row:" + str(self.line_num) + "] " + \
                   self.t_type + ": " + \
                   self.format_quantity(self.buy_quantity) + " " + \
                   self.buy_asset + \
                   self.format_value(self.buy_value) + " \"" + \
                   self.wallet + "\" " + \
                   self.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z')
        elif self.t_type in self.SELL_TYPES or self.t_type == self.TYPE_WITHDRAWAL:
            return "[Row:" + str(self.line_num) + "] " + \
                   self.t_type + ": " + \
                   self.format_quantity(self.sell_quantity) + " " + \
                   self.sell_asset + \
                   self.format_value(self.sell_value) + " \"" + \
                   self.wallet + "\" " + \
                   self.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z')
        else:
            return "[Row:" + str(self.line_num) + "] " + \
                   self.t_type + ": " + \
                   self.format_quantity(self.buy_quantity) + " " + \
                   self.buy_asset + \
                   self.format_value(self.buy_value) + " <- " + \
                   self.format_quantity(self.sell_quantity) + " " + \
                   self.sell_asset + \
                   self.format_value(self.sell_value) + " \"" + \
                   self.wallet + "\" " + \
                   self.timestamp.strftime('%Y-%m-%dT%H:%M:%S %Z')

class TransactionInRecord(TransactionRecordBase):
    def validate(self):
        if self.t_type in self.BUY_TYPES or self.t_type == self.TYPE_DEPOSIT:
            if not self.buy_asset or self.buy_quantity is None:
                raise Exception("Type: " + self.t_type + " missing buy details")
            if self.sell_asset or self.sell_quantity is not None:
                raise Exception("Type: " + self.t_type + " sell details not required")
            if self.buy_quantity < 0:
                raise Exception("Type: " + self.t_type + " buy quantity is negative")
        elif self.t_type in self.SELL_TYPES or self.t_type == self.TYPE_WITHDRAWAL:
            if not self.sell_asset or self.sell_quantity is None:
                raise Exception("Type: " + self.t_type + " missing sell details")
            elif self.buy_asset or self.buy_quantity is not None:
                raise Exception("Type: " + self.t_type + " buy details not required")
            elif self.sell_quantity < 0:
                raise Exception("Type: " + self.t_type + " sell quantity is negative")
        elif self.t_type == self.TYPE_TRADE:
            if not self.buy_asset or self.buy_quantity is None:
                raise Exception("Type: " + self.t_type + " missing buy details")
            if not self.sell_asset or self.sell_quantity is None:
                raise Exception("Type: " + self.t_type + " missing sell details")
            if self.buy_quantity < 0:
                raise Exception("Type: " + self.t_type + " buy quantity is negative")
            if self.sell_quantity < 0:
                raise Exception("Type: " + self.t_type + " sell quantity is negative")
        else:
            raise ValueError("Type: " + self.t_type + " is unrecognised")

        if self.fee_asset and (self.fee_asset != self.buy_asset
                               and self.fee_asset != self.sell_asset):
            raise Exception("Fee asset: " + self.fee_asset + " does not match")
        elif self.fee_asset and self.fee_quantity is None:
            raise Exception("Missing fee quantity")
        elif self.fee_asset and self.fee_quantity < 0:
            raise Exception("Fee quantity is negative")

    def normalise_to_localtime(self):
        self.timestamp = self.timestamp.astimezone(config.TZ_LOCAL)

    def include_fees(self):
        # Include fees within buy/sell portion
        if self.buy_asset == self.fee_asset and self.fee_quantity:
            self.buy_quantity -= self.fee_quantity
            if self.fee_value:
                self.buy_value -= self.fee_value

            self.fee_quantity = None
            self.fee_asset = ""
            self.fee_value = None
        elif self.sell_asset == self.fee_asset and self.fee_quantity:
            self.sell_quantity += self.fee_quantity
            if self.fee_value:
                self.sell_value += self.fee_value

            self.fee_quantity = None
            self.fee_asset = ""
            self.fee_value = None



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Coinbase"
DUPLICATE = "Duplicate"

def parse_coinbase_transfers(in_row):
    if in_row[1] == "Deposit":
        return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[5],
                                    buy_asset=in_row[6],
                                    fee_quantity=in_row[4],
                                    fee_asset=in_row[6],
                                    wallet=WALLET)
    elif in_row[1] == "Withdrawal":
        return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                    DataParser.parse_timestamp(in_row[0]),
                                    sell_quantity=in_row[5],
                                    sell_asset=in_row[6],
                                    fee_quantity=in_row[4],
                                    fee_asset=in_row[6],
                                    wallet=WALLET)
    elif in_row[1] == "Buy":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[2],
                                    buy_asset="BTC",
                                    sell_quantity=in_row[3],
                                    sell_asset=in_row[6],
                                    fee_quantity=in_row[4],
                                    fee_asset=in_row[6],
                                    wallet=WALLET)
    elif in_row[1] == "Sell":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[3],
                                    buy_asset=in_row[6],
                                    sell_quantity=in_row[2],
                                    sell_asset="BTC",
                                    fee_quantity=in_row[4],
                                    fee_asset=in_row[6],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised Type: " + in_row[1])

def parse_coinbase_transactions(in_row):
    if in_row[21] != "":
        # Hash so must be external crypto deposit or withdrawal
        if Decimal(in_row[2]) < 0:
            return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                        DataParser.parse_timestamp(in_row[0]),
                                        sell_quantity=abs(Decimal(in_row[2])),
                                        sell_asset=in_row[3],
                                        wallet=WALLET)
        else:
            return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                        DataParser.parse_timestamp(in_row[0]),
                                        buy_quantity=in_row[2],
                                        buy_asset=in_row[3],
                                        wallet=WALLET)
    elif in_row[12] != "":
        # Transfer ID so could be a trade or external fiat deposit/withdrawal
        if in_row[3] != in_row[8]:
            # Currencies are different so must be a trade
            if Decimal(in_row[2]) < 0:
                return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                            DataParser.parse_timestamp(in_row[0]),
                                            buy_quantity=Decimal(in_row[7]) + Decimal(in_row[9]),
                                            buy_asset=in_row[8],
                                            sell_quantity=abs(Decimal(in_row[2])),
                                            sell_asset=in_row[3],
                                            fee_quantity=in_row[9],
                                            fee_asset=in_row[10],
                                            wallet=WALLET)
            else:
                return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                            DataParser.parse_timestamp(in_row[0]),
                                            buy_quantity=in_row[2],
                                            buy_asset=in_row[3],
                                            sell_quantity=Decimal(in_row[7]) - Decimal(in_row[9]),
                                            sell_asset=in_row[8],
                                            fee_quantity=in_row[9],
                                            fee_asset=in_row[10],
                                            wallet=WALLET)
        else:
            if Decimal(in_row[2]) < 0:
                return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                            DataParser.parse_timestamp(in_row[0]),
                                            sell_quantity=in_row[7],
                                            sell_asset=in_row[3],
                                            fee_quantity=in_row[9],
                                            fee_asset=in_row[10],
                                            wallet=WALLET)
            else:
                return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                            DataParser.parse_timestamp(in_row[0]),
                                            buy_quantity=in_row[7],
                                            buy_asset=in_row[3],
                                            fee_quantity=in_row[9],
                                            fee_asset=in_row[10],
                                            wallet=WALLET)
    else:
        # Could be a referral bonus or deposit/withdrawal to/from Coinbase Pro
        if in_row[5] != "" and in_row[3] == "BTC":
            # Bonus is always in BTC
            return TransactionOutRecord(TransactionOutRecord.TYPE_GIFT_RECEIVED,
                                        DataParser.parse_timestamp(in_row[0]),
                                        buy_quantity=in_row[2],
                                        buy_asset=in_row[3],
                                        wallet=WALLET)
        elif in_row[5] != "" and in_row[3] != "BTC":
            # Special case, flag as duplicate entry, trade will be in BTC Wallet Transactions Report
            if Decimal(in_row[2]) < 0:
                return TransactionOutRecord(DUPLICATE,
                                            DataParser.parse_timestamp(in_row[0]),
                                            sell_quantity=abs(Decimal(in_row[2])),
                                            sell_asset=in_row[3],
                                            wallet=WALLET)
            else:
                return TransactionOutRecord(DUPLICATE,
                                            DataParser.parse_timestamp(in_row[0]),
                                            buy_quantity=in_row[2],
                                            buy_asset=in_row[3],
                                            wallet=WALLET)
        elif Decimal(in_row[2]) < 0:
            return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                        DataParser.parse_timestamp(in_row[0]),
                                        sell_quantity=abs(Decimal(in_row[2])),
                                        sell_asset=in_row[3],
                                        wallet=WALLET)
        else:
            return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                        DataParser.parse_timestamp(in_row[0]),
                                        buy_quantity=in_row[2],
                                        buy_asset=in_row[3],
                                        wallet=WALLET)

DataParser(DataParser.TYPE_EXCHANGE,
           "Coinbase Transfers",
           ['Timestamp', 'Type', None, 'Subtotal', 'Fees', 'Total', 'Currency', 'Price Per Coin',
            'Payment Method', 'ID', 'Share'],
           row_handler=parse_coinbase_transfers)

DataParser(DataParser.TYPE_EXCHANGE,
           "Coinbase Transactions",
           ['Timestamp', 'Balance', 'Amount', 'Currency', 'To', 'Notes', 'Instantly Exchanged',
            'Transfer Total', 'Transfer Total Currency', 'Transfer Fee', 'Transfer Fee Currency',
            'Transfer Payment Method', 'Transfer ID', 'Order Price', 'Order Currency', 'Order BTC',
            'Order Tracking Code', 'Order Custom Parameter', 'Order Paid Out',
            'Recurring Payment ID', None, None],
           row_handler=parse_coinbase_transactions)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Ethereum"

def parse_etherscan(in_row):
    if Decimal(in_row[7]) > 0:
        return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                    DataParser.parse_timestamp(int(in_row[2])),
                                    buy_quantity=in_row[7],
                                    buy_asset="ETH",
                                    wallet=WALLET)
    elif Decimal(in_row[8]) > 0:
        return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                    DataParser.parse_timestamp(int(in_row[2])),
                                    sell_quantity=in_row[8],
                                    sell_asset="ETH",
                                    fee_quantity=in_row[10],
                                    fee_asset="ETH",
                                    wallet=WALLET)
    else:
        return TransactionOutRecord(TransactionOutRecord.TYPE_SPEND,
                                    DataParser.parse_timestamp(int(in_row[2])),
                                    sell_quantity=in_row[8],
                                    sell_asset="ETH",
                                    fee_quantity=in_row[10],
                                    fee_asset="ETH",
                                    wallet=WALLET)

DataParser(DataParser.TYPE_EXPLORER,
           "Etherscan (Ethereum)",
           ['Txhash', 'Blockno', 'UnixTimestamp', 'DateTime', 'From', 'To', 'ContractAddress',
            'Value_IN(ETH)', 'Value_OUT(ETH)', None, 'TxnFee(ETH)', 'TxnFee(USD)',
            'Historical $Price/Eth', 'Status', 'ErrCode'],
           row_handler=parse_etherscan)

DataParser(DataParser.TYPE_EXPLORER,
           "Etherscan (Ethereum)",
           ['Txhash', 'Blockno', 'UnixTimestamp', 'DateTime', 'From', 'To', 'ContractAddress',
            'Value_IN(ETH)', 'Value_OUT(ETH)', None, 'TxnFee(ETH)', 'TxnFee(USD)',
            'Historical $Price/Eth', 'Status', 'ErrCode', 'PrivateNote'],
           row_handler=parse_etherscan)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Bifinex"

def parse_bitfinex_trades(in_row):
    if Decimal(in_row[2]) > 0:
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[6], dayfirst=True),
                                    buy_quantity=in_row[2],
                                    buy_asset=in_row[1].split('/')[0],
                                    sell_quantity=Decimal(in_row[3]) * Decimal(in_row[2]),
                                    sell_asset=in_row[1].split('/')[1],
                                    fee_quantity=abs(Decimal(in_row[4])),
                                    fee_asset=in_row[5],
                                    wallet=WALLET)
    else:
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[6], dayfirst=True),
                                    buy_quantity=Decimal(in_row[3]) * abs(Decimal(in_row[2])),
                                    buy_asset=in_row[1].split('/')[1],
                                    sell_quantity=abs(Decimal(in_row[2])),
                                    sell_asset=in_row[1].split('/')[0],
                                    fee_quantity=abs(Decimal(in_row[4])),
                                    fee_asset=in_row[5],
                                    wallet=WALLET)

def parse_bitfinex_deposits_withdrawals(in_row):
    if in_row[3] != "COMPLETED":
        return None

    if Decimal(in_row[4]) > 0:
        return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                    DataParser.parse_timestamp(in_row[1], dayfirst=True),
                                    buy_quantity=in_row[4],
                                    buy_asset=in_row[2],
                                    fee_quantity=abs(Decimal(in_row[5])),
                                    fee_asset=in_row[2],
                                    wallet=WALLET)

    else:
        return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                    DataParser.parse_timestamp(in_row[1], dayfirst=True),
                                    sell_quantity=abs(Decimal(in_row[4])),
                                    sell_asset=in_row[2],
                                    fee_quantity=abs(Decimal(in_row[5])),
                                    fee_asset=in_row[2],
                                    wallet=WALLET)

DataParser(DataParser.TYPE_EXCHANGE,
           "Bitfinex Trades",
           ['#', 'PAIR', 'AMOUNT', 'PRICE', 'FEE', 'FEE CURRENCY', 'DATE', 'ORDER ID'],
           row_handler=parse_bitfinex_trades)

DataParser(DataParser.TYPE_EXCHANGE,
           "Bitfinex Deposits/Withdrawals",
           ['#', 'DATE', 'CURRENCY', 'STATUS', 'AMOUNT', 'FEES', 'DESCRIPTION', 'TRANSACTION ID'],
           row_handler=parse_bitfinex_deposits_withdrawals)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Wirex"

def parse_wirex(in_row):
    if in_row[1] == "Create":
        return None
    elif in_row[1] == "In":
        return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                    DataParser.parse_timestamp(in_row[2]),
                                    buy_quantity=in_row[3].split(' ')[0],
                                    buy_asset=in_row[3].split(' ')[1],
                                    wallet=WALLET)
    elif in_row[1] == "Out":
        return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                    DataParser.parse_timestamp(in_row[2]),
                                    sell_quantity=in_row[3].split(' ')[0],
                                    sell_asset=in_row[3].split(' ')[1],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised Type: " + in_row[1])

DataParser(DataParser.TYPE_EXCHANGE,
           "Wirex",
           ['# ', None, 'Time ', 'Amount', 'Available'],
           row_handler=parse_wirex)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal

from ...config import config
from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Electrum"

def parse_electrum(in_row):
    if not config.args.cryptoasset:
        raise Exception(config.ERROR_TXT[0])

    if Decimal(in_row[3]) > 0:
        return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                    DataParser.parse_timestamp(in_row[4]),
                                    buy_quantity=Decimal(in_row[3]),
                                    buy_asset=config.args.cryptoasset,
                                    wallet=WALLET)

    else:
        return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                    DataParser.parse_timestamp(in_row[4]),
                                    sell_quantity=abs(Decimal(in_row[3])),
                                    sell_asset=config.args.cryptoasset,
                                    wallet=WALLET)

DataParser(DataParser.TYPE_WALLET,
           "Electrum",
           ['transaction_hash', 'label', 'confirmations', 'value', 'timestamp'],
           row_handler=parse_electrum)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging

from .config import config

log = logging.getLogger()

class Wallet(object):
    wallets = {}

    @classmethod
    def add_tokens(cls, wallet, asset, quantity):
        if (wallet, asset) not in cls.wallets:
            cls.wallets[(wallet, asset)] = Wallet(wallet, asset)

        cls.wallets[(wallet, asset)].add(quantity)

    @classmethod
    def subtract_tokens(cls, wallet, asset, quantity):
        if (wallet, asset) not in cls.wallets:
            cls.wallets[(wallet, asset)] = Wallet(wallet, asset)

        cls.wallets[(wallet, asset)].subtract(quantity)

    def __init__(self, wallet, asset):
        self.wallet = wallet
        self.asset = asset
        self.balance = 0

    def _format_balance(self):
        return '{:0,f}'.format(self.balance.normalize())

    def _format_wallet_name(self):
        return self.wallet + ":" + self.asset

    def add(self, quantity):
        self.balance += quantity
        log.debug("%s=%s (+%s)",
                  self._format_wallet_name(),
                  self._format_balance(),
                  '{:0,f}'.format(quantity))

    def subtract(self, quantity):
        self.balance -= quantity
        log.debug("%s=%s (-%s)",
                  self._format_wallet_name(),
                  self._format_balance(),
                  '{:0,f}'.format(quantity))

        if self.balance < 0 and self.asset not in config.fiat_list:
            log.warning("Balance at %s is negative %s",
                        self._format_wallet_name(),
                        self._format_balance())

    def __str__(self):
        return self._format_wallet_name() + "=" + self._format_balance()

def audit_transactions(transaction_records):
    log.debug("==FULL AUDIT TRANSACTIONS==")
    for tr in transaction_records:
        if tr.buy_asset:
            Wallet.add_tokens(tr.wallet, tr.buy_asset, tr.buy_quantity)

        if tr.sell_asset:
            Wallet.subtract_tokens(tr.wallet, tr.sell_asset, tr.sell_quantity)

    log.info("==FINAL AUDIT BALANCES==")
    for w in sorted(Wallet.wallets):
        log.info(Wallet.wallets[w])



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Binance"

def parse_binance_trades(in_row):
    if in_row[2] == "BUY":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[4],
                                    buy_asset=in_row[7],
                                    sell_quantity=in_row[5],
                                    sell_asset=in_row[1].replace(in_row[7], ''),
                                    fee_quantity=in_row[6],
                                    fee_asset=in_row[7],
                                    wallet=WALLET)
    elif in_row[2] == "SELL":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[5],
                                    buy_asset=in_row[7],
                                    sell_quantity=in_row[4],
                                    sell_asset=in_row[1].replace(in_row[7], ''),
                                    fee_quantity=in_row[6],
                                    fee_asset=in_row[7],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised Type: " + in_row[2])

def parse_binance_deposits_withdrawals(in_row):
    if in_row[8] != "Completed":
        return None

    # Assume that a transaction fee of 0 must be a Deposit
    if Decimal(in_row[3]) == 0:
        return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=in_row[2],
                                    buy_asset=in_row[1],
                                    wallet=WALLET)
    else:
        return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                    DataParser.parse_timestamp(in_row[0]),
                                    sell_quantity=in_row[2],
                                    sell_asset=in_row[1],
                                    fee_quantity=in_row[3],
                                    fee_asset=in_row[1],
                                    wallet=WALLET)

DataParser(DataParser.TYPE_EXCHANGE,
           "Binance Trades",
           ['Date(UTC)', 'Market', 'Type', 'Price', 'Amount', 'Total', 'Fee', 'Fee Coin'],
           row_handler=parse_binance_trades)

DataParser(DataParser.TYPE_EXCHANGE,
           "Binance Deposits/Withdrawals",
           ['Date', 'Coin', 'Amount', 'TransactionFee', 'Address', 'TXID', 'SourceAddress',
            'PaymentID', 'Status'],
           row_handler=parse_binance_deposits_withdrawals)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Ledger Live"

def parse_ledger_live(in_row):
    if in_row[2] == "IN":
        return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                    DataParser.parse_timestamp(in_row[0]),
                                    buy_quantity=Decimal(in_row[3]) + Decimal(in_row[4]),
                                    buy_asset=in_row[1],
                                    fee_quantity=Decimal(in_row[4]),
                                    fee_asset=in_row[1],
                                    wallet=WALLET)

    if in_row[2] == "OUT":
        return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                    DataParser.parse_timestamp(in_row[0]),
                                    sell_quantity=Decimal(in_row[3]) - Decimal(in_row[4]),
                                    sell_asset=in_row[1],
                                    fee_quantity=Decimal(in_row[4]),
                                    fee_asset=in_row[1],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised Operation Type: " + in_row[2])

DataParser(DataParser.TYPE_WALLET,
           "Ledger Live",
           ['Operation Date', 'Currency Ticker', 'Operation Type', 'Operation Amount',
            'Operation Fees', 'Operation Hash', 'Account Name', 'Account id'],
           row_handler=parse_ledger_live)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from decimal import Decimal

import dateutil.tz

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "OKEx"
TZ_INFOS = {'CST': dateutil.tz.gettz('Asia/Shanghai')}

def parse_okex_trades(all_in_row):
    t_records = []

    for in_row in all_in_row:
        if in_row[1] == "buy":
            buy_quantity = in_row[2]
            buy_asset = in_row[5]
            fee_quantity = abs(Decimal(in_row[4]))
            fee_asset = in_row[5]

            t_records.append(None)
        elif in_row[1] == "sell":
            sell_quantity = abs(Decimal(in_row[2]))
            sell_asset = in_row[5]

            t_record = TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                            DataParser.parse_timestamp(in_row[0],
                                                                       tzinfos=TZ_INFOS),
                                            buy_quantity=buy_quantity,
                                            buy_asset=buy_asset,
                                            sell_quantity=sell_quantity,
                                            sell_asset=sell_asset,
                                            fee_quantity=fee_quantity,
                                            fee_asset=fee_asset,
                                            wallet=WALLET)
            t_records.append(t_record)
        else:
            raise ValueError("Unrecognised type: " + in_row[1])

    return t_records

DataParser(DataParser.TYPE_EXCHANGE,
           "OKEx",
           ['time', 'type', 'size', 'balance', 'fee', 'currency'],
           all_handler=parse_okex_trades)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Bittrex"

def parse_bittrex_trades2(in_row):
    if in_row[3] == "LIMIT_BUY":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[14]),
                                    buy_quantity=in_row[5],
                                    buy_asset=in_row[1].split('-')[1],
                                    sell_quantity=in_row[8],
                                    sell_asset=in_row[1].split('-')[0],
                                    fee_quantity=in_row[7],
                                    fee_asset=in_row[1].split('-')[0],
                                    wallet=WALLET)
    elif in_row[3] == "LIMIT_SELL":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[14]),
                                    buy_quantity=in_row[8],
                                    buy_asset=in_row[1].split('-')[0],
                                    sell_quantity=in_row[5],
                                    sell_asset=in_row[1].split('-')[1],
                                    fee_quantity=in_row[7],
                                    fee_asset=in_row[1].split('-')[0],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised Order Type: " + in_row[2])

def parse_bittrex_trades(in_row):
    if in_row[2] == "LIMIT_BUY":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[8]),
                                    buy_quantity=in_row[3],
                                    buy_asset=in_row[1].split('-')[1],
                                    sell_quantity=in_row[6],
                                    sell_asset=in_row[1].split('-')[0],
                                    fee_quantity=in_row[5],
                                    fee_asset=in_row[1].split('-')[0],
                                    wallet=WALLET)
    elif in_row[2] == "LIMIT_SELL":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[8]),
                                    buy_quantity=in_row[6],
                                    buy_asset=in_row[1].split('-')[0],
                                    sell_quantity=in_row[3],
                                    sell_asset=in_row[1].split('-')[1],
                                    fee_quantity=in_row[5],
                                    fee_asset=in_row[1].split('-')[0],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised Order Type: " + in_row[2])

def parse_bittrex_deposits(in_row):
    return TransactionOutRecord(TransactionOutRecord.TYPE_DEPOSIT,
                                DataParser.parse_timestamp(in_row[4]),
                                buy_quantity=in_row[1],
                                buy_asset=in_row[2],
                                wallet=WALLET)

def parse_bittrex_withdrawals(in_row):
    return TransactionOutRecord(TransactionOutRecord.TYPE_WITHDRAWAL,
                                DataParser.parse_timestamp(in_row[4]),
                                sell_quantity=in_row[2],
                                sell_asset=in_row[1],
                                fee_quantity=in_row[7],
                                fee_asset=in_row[1],
                                wallet=WALLET)

DataParser(DataParser.TYPE_EXCHANGE,
           "Bittrex Trades",
           ['Uuid', 'Exchange', 'TimeStamp', 'OrderType', 'Limit', 'Quantity', 'QuantityRemaining',
            'Commission', 'Price', 'PricePerUnit', 'IsConditional', 'Condition', 'ConditionTarget',
            'ImmediateOrCancel', 'Closed'],
           row_handler=parse_bittrex_trades2)

DataParser(DataParser.TYPE_EXCHANGE,
           "Bittrex Trades",
           ['OrderUuid', 'Exchange', 'Type', 'Quantity', 'Limit', 'CommissionPaid', 'Price',
            'Opened', 'Closed'],
           row_handler=parse_bittrex_trades)

DataParser(DataParser.TYPE_EXCHANGE,
           "Bittrex Deposits",
           ['Id', 'Amount', 'Currency', 'Confirmations', 'LastUpdated', 'TxId', 'CryptoAddress'],
           row_handler=parse_bittrex_deposits)

DataParser(DataParser.TYPE_EXCHANGE,
           "Bittrex Withdrawals",
           ['PaymentUuid', 'Currency', 'Amount', 'Address', 'Opened', 'Authorized',
            'PendingPayment', 'TxCost', 'TxId', 'Canceled', 'InvalidAddress'],
           row_handler=parse_bittrex_withdrawals)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

from ..out_record import TransactionOutRecord
from ..dataparser import DataParser

WALLET = "Crypsty"

def parse_cryptsy(in_row):
    if in_row[1] == "Buy":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[8], tz='US/Eastern'),
                                    buy_quantity=in_row[4],
                                    buy_asset=in_row[2].split('/')[0],
                                    sell_quantity=in_row[7],
                                    sell_asset=in_row[2].split('/')[1],
                                    fee_quantity=in_row[6],
                                    fee_asset=in_row[2].split('/')[1],
                                    wallet=WALLET)
    elif in_row[1] == "Sell":
        return TransactionOutRecord(TransactionOutRecord.TYPE_TRADE,
                                    DataParser.parse_timestamp(in_row[8], tz='US/Eastern'),
                                    buy_quantity=in_row[7],
                                    buy_asset=in_row[2].split('/')[1],
                                    sell_quantity=in_row[4],
                                    sell_asset=in_row[2].split('/')[0],
                                    fee_quantity=in_row[6],
                                    fee_asset=in_row[2].split('/')[1],
                                    wallet=WALLET)
    else:
        raise ValueError("Unrecognised OrderType: " + in_row[1])

DataParser(DataParser.TYPE_EXCHANGE,
           "Cryptsy",
           ['TradeID', 'OrderType', 'Market', 'Price', 'Quantity', 'Total', 'Fee', 'Net',
            'Timestamp'],
           row_handler=parse_cryptsy)



# -*- coding: utf-8 -*-
# (c) Nano Nano Ltd 2019

import logging
from decimal import Decimal
from datetime import datetime

from ..version import __version__
from ..config import config
from .pricedata import PriceData

log = logging.getLogger()

class ValueAsset(object):
    def __init__(self):
        self.price_data = PriceData()

    def get_value(self, asset, timestamp, quantity, fixed_value):
        if asset == config.CCY:
            return quantity

        if fixed_value is None:
            price_ccy, _, _ = self.get_historical_price(asset, timestamp)
            if price_ccy is not None:
                value = price_ccy * quantity
                log.debug("Price on %s, 1 %s=%s%s %s, %s %s=%s%s %s",
                          timestamp.strftime('%Y-%m-%d'),
                          asset,
                          config.sym(), '{:0,.2f}'.format(price_ccy), config.CCY,
                          str(quantity),
                          asset,
                          config.sym(), '{:0,.2f}'.format(value), config.CCY)
            else:
                value = Decimal(0)
                log.warning("Price at %s for %s is not available",
                            timestamp.strftime('%Y-%m-%d'),
                            asset)
        else:
            value = fixed_value
            log.debug("Using fixed value, %s %s=%s%s %s",
                      str(quantity),
                      asset,
                      config.sym(), '{:0,.2f}'.format(value), config.CCY)

        return value

    def get_current_value(self, asset, quantity):
        price_ccy, name, data_source = self.get_latest_price(asset)
        if price_ccy is not None:
            return price_ccy * quantity, name, data_source

        return Decimal(0), None, None

    def get_historical_price(self, asset, timestamp):
        if timestamp.date() >= datetime.now().date():
            raise Exception("Date is not in the past: " + timestamp.strftime('%Y-%m-%d'))

        if asset == "BTC" or asset in config.fiat_list:
            price_ccy, name, data_source = self.price_data.get_historical(asset,
                                                                          config.CCY,
                                                                          timestamp)
        else:
            price_btc, _, _ = self.price_data.get_historical("BTC",
                                                             config.CCY,
                                                             timestamp)
            if price_btc is not None:
                price_ccy, name, data_source = self.price_data.get_historical(asset,
                                                                              "BTC",
                                                                              timestamp)
                if price_ccy is not None:
                    price_ccy = price_btc * price_ccy

        return price_ccy, name, data_source

    def get_latest_price(self, asset):
        if asset == "BTC" or asset in config.fiat_list:
            price_ccy, name, data_source = self.price_data.get_latest(asset, config.CCY)
        else:
            price_btc, _, _ = self.price_data.get_latest("BTC", config.CCY)

            if price_btc is not None:
                price_ccy, name, data_source = self.price_data.get_latest(asset, "BTC")
                if price_ccy is not None:
                    price_ccy = price_btc * price_ccy

        return price_ccy, name, data_source
